#!/bin/bash
#
# Copyright (c) 2019-2020 Oracle and/or its affiliates. All rights reserved.
# Licensed under the GNU General Public License Version 2 as shown at https://oss.oracle.com/licenses/GPL-2

VAULT_TOKEN=""
VAULT_CERT_SANS=""
VAULT_ADDRESS=""
VAULT_SKIP_VERIFY="false"
FORCE_DOWNLOAD_CERTS="false"
OLCNE_SECRET_MANAGER_TYPE=""
OLCNE_CA_PATH=""
OLCNE_NODE_CERT_PATH=""
OLCNE_NODE_KEY_PATH=""
OLCNE_COMPONENT="agent"
OLCNE_TLS_MIN_VERSION=""
OLCNE_TLS_CIPHER_SUITES=""
OLCNE_DEBUG="0"

function get_cmd {
    flag=$1
    argument=$2

    if [[ "${argument}" != "''" ]]; then
        echo "${flag} ${argument}"
    fi
}

function get_environment_unit {
    binary_name=$1
    extra_args=$2

    vault_token_cmd=$(get_cmd --vault-token "'${VAULT_TOKEN}'")
    secret_manager_type_cmd=$(get_cmd --secret-manager-type "'${OLCNE_SECRET_MANAGER_TYPE}'")
    vault_address_cmd=$(get_cmd --vault-address "'${VAULT_ADDRESS}'")
    vault_cert_sans_cmd=$(get_cmd --vault-cert-sans "'${VAULT_CERT_SANS}'")
    olcne_ca_path_cmd=$(get_cmd --olcne-ca-path "'${OLCNE_CA_PATH}'")
    olcne_node_cert_path_cmd=$(get_cmd --olcne-node-cert-path "'${OLCNE_NODE_CERT_PATH}'")
    olcne_node_key_path_cmd=$(get_cmd --olcne-node-key-path "'${OLCNE_NODE_KEY_PATH}'")
    olcne_tls_min_version=$(get_cmd --olcne-tls-min-version "'${OLCNE_TLS_MIN_VERSION}'")
    olcne_tls_cipher_suites=$(get_cmd --olcne-tls-cipher-suites "'${OLCNE_TLS_CIPHER_SUITES}'")
    echo "
[Service]
Environment=\"VAULT_SKIP_VERIFY=${VAULT_SKIP_VERIFY}\"
Environment=\"OLCNE_DEBUG=${OLCNE_DEBUG}\"
Environment=\"OLCNE_AGENT_ARGS=${vault_token_cmd} ${secret_manager_type_cmd} ${vault_address_cmd} ${vault_cert_sans_cmd} ${olcne_ca_path_cmd} ${olcne_node_cert_path_cmd} ${olcne_node_key_path_cmd}\"
Environment=\"OLCNE_TLS= ${olcne_tls_min_version} ${olcne_tls_cipher_suites}\"
ExecStart=
ExecStart=/usr/libexec/${binary_name} ${extra_args} \$OLCNE_AGENT_ARGS \$OLCNE_TLS
RestartSec=5
"
}

function usage {
	echo "The Platform Bootstrap tool is used to configure Vault for the Oracle Linux Cloud Native Environment" >&2
    echo ""
    echo "-t | --vault-token <VAULT_TOKEN>"
    echo "      Token to talk with vault"
    echo "-s | --vault-cert-sans <VAULT_CERT_SANS>"
    echo "      The cert sans you would like the Platform Certificate to have."
	echo "-A | --vault-address <VAULT_ADDRESS>"
    echo "      Specify the address for Vault. Mostly used along with '--production' to setup a vault."
    echo "-S | --vault-skip-verify"
    echo "      Skip cert CA verification. DEFAULT: ${VAULT_SKIP_VERIFY}"
    echo "-f | --force-download-certs"
    echo "      Remove certs before restarting the systemd service. DEFAULT: ${FORCE_DOWNLOAD_CERTS}"
    echo "-T | --secret-manager-type <OLCNE_SECRET_MANAGER_TYPE>"
    echo "      Platform secret manager type (file, vault). DEFAULT: ${OLCNE_SECRET_MANAGER_TYPE}"
    echo "-c | --olcne-ca-path <OLCNE_CA_PATH>"
    echo "      Path to your Platform CA. In the case of vault this is where it will write the CA."
    echo "-C | --olcne-node-cert-path <OLCNE_CERT_PATH>"
    echo "      Path to your Platform Certificate In the case of vault this is where it will write the cert."
    echo "-K | --olcne-node-key-path <OLCNE_KEY_PATH>"
    echo "      Path to your Platform Key In the case of vault this is where it will write the key."
    echo "--olcne-tls-min-version <OLCNE_TLS_MIN_VERSION>"
    echo "	TLS Minimum version."
    echo "--olcne-tls-cipher-suites <OLCNE_TLS_CIPHER_SUITES>"
    echo "	TLS Cipher Suites."
    echo "-o | --olcne-component <OLCNE_COMPONENT>"
    echo "      Systemd component to start. (agent, api-server) DEFAULT: ${OLCNE_COMPONENT}"
	echo "-d | --debug"
    echo "      Start the olcne systemd service with debug enabled"
    echo "usage: " >&2
	echo "   $0 [ --vault-token <VAULT_TOKEN> --vault-cert-sans <VAULT_CERT_SANS> --vault-address <VAULT_ADDRESS>
                  --secret-manager-type <OLCNE_SECRET_MANAGER_TYPE> --olcne-ca-path <OLCNE_CA_PATH> --olcne-node-cert-path <OLCNE_CERT_PATH>
                  --olcne-node-key-path <OLCNE_KEY_PATH> --olcne-tls-min-version <OLCNE_TLS_MIN_VERSION> --olcne-tls-cipher-suites <OLCNE_TLS_CIPHER_SUITES>
		  --olcne-component <OLCNE_COMPONENT> --vault-skip-verify --force-download-certs --help ]" >&2
	exit 0
}

function main {
    # Validate component
    if [[ ${OLCNE_COMPONENT} != "agent" && ${OLCNE_COMPONENT} != "api-server" ]]; then
        echo "Invalid Platform Component ${OLCNE_COMPONENT}. Expected 'agent' or 'api-server'"
        exit 1
    fi

    if [[ ${FORCE_DOWNLOAD_CERTS} == "true" ]]; then
        rm -rf "${OLCNE_CA_PATH}"
        rm -rf "${OLCNE_NODE_CERT_PATH}"
        rm -rf "${OLCNE_NODE_KEY_PATH}"
    fi

    if [[ ${OLCNE_COMPONENT} == "agent" ]]; then
        # Create directory for unit files
        mkdir -p /etc/systemd/system/olcne-agent.service.d

        get_environment_unit olcne-agent > /etc/systemd/system/olcne-agent.service.d/10-auth.conf

        systemctl daemon-reload
        sleep 1
        systemctl enable olcne-agent
        sleep 1
        systemctl restart olcne-agent
        sleep 2
        systemctl status olcne-agent --no-pager
    fi

    if [[ ${OLCNE_COMPONENT} == "api-server" ]]; then
        # Create directory for unit files
        mkdir -p /etc/systemd/system/olcne-api-server.service.d

       get_environment_unit olcne-api-server '-i /etc/olcne/modules' > /etc/systemd/system/olcne-api-server.service.d/10-auth.conf

        systemctl daemon-reload
        sleep 1
        systemctl enable olcne-api-server
        sleep 1
        systemctl restart olcne-api-server
        sleep 2
        systemctl status olcne-api-server --no-pager
    fi
}

# MAIN
while [[ $# -gt 0 ]]; do
key="$1"

case $key in
    -t | --vault-token )
        VAULT_TOKEN="$2"
        shift # shift past argument
        shift # shift past value
    ;;
   -s | --vault-cert-sans )
        VAULT_CERT_SANS="$2"
        shift # shift past argument
        shift # shift past value
    ;;
    -A | --vault-address )
        VAULT_ADDRESS="$2"
        shift # shift past argument
        shift # shift past value
    ;;
    -S | --vault-skip-verify )
        VAULT_SKIP_VERIFY="true"
        shift # shift past argument
    ;;
    -f | --force-download-certs )
        FORCE_DOWNLOAD_CERTS="true"
        shift # shift past argument
    ;;
    -T | --secret-manager-type )
        OLCNE_SECRET_MANAGER_TYPE="$2"
        shift # shift past argument
        shift # shift past value
    ;;
    -c | --olcne-ca-path )
        OLCNE_CA_PATH="$2"
        shift # shift past argument
        shift # shift past value
    ;;
    -C | --olcne-node-cert-path )
        OLCNE_NODE_CERT_PATH="$2"
        shift # shift past argument
        shift # shift past value
    ;;
    -K | --olcne-node-key-path )
        OLCNE_NODE_KEY_PATH="$2"
        shift # shift past argument
        shift # shift past value
    ;;
    --olcne-tls-min-version )
	OLCNE_TLS_MIN_VERSION="$2"
	shift # shift past argument
	shift # shift past value
    ;;
    --olcne-tls-cipher-suites )
	OLCNE_TLS_CIPHER_SUITES="$2"
	shift # shift past argument
	shift # shift past value
    ;;
    -o | --olcne-component )
        OLCNE_COMPONENT="$2"
        shift # shift past argument
        shift # shift past value
    ;;
    -d | --debug )
        OLCNE_DEBUG="1"
        shift # shift past argument
    ;;
    -h | --help )
        usage
        shift # shift past argument
    ;;
    *) # unknown option
        echo "[ERROR] unknown argument: ${key}"
        usage
        exit 1
    ;;
esac
done

main
