#!/bin/bash
#
# Copyright (c) 2019-2020 Oracle and/or its affiliates. All rights reserved.
# Licensed under the GNU General Public License Version 2 as shown at https://oss.oracle.com/licenses/GPL-2

# USER INPUT VARS
# Used to skip overwriting the vault systemd and vault config if set to true
PRODUCTION="false"
VAULT_ADDR="https://127.0.0.1:8200"
# Directory to store all vault data
VAULT_DATA_DIR="/var/lib/vault.d/data"
VAULT_SERVER_CA_DIR="/etc/vault.d/certificates"
VAULT_SERVER_EXTRA_CERT_SANS="localhost,127.0.0.1"
VAULT_SERVER_TLS_CERT_FILE="${VAULT_SERVER_CA_DIR}/vault_server.cert"
VAULT_SERVER_TLS_KEY_FILE="${VAULT_SERVER_CA_DIR}/vault_server.key"

VAULT_SERVER_TLS_CA_KEY_FILE="${VAULT_SERVER_CA_DIR}/ca.key"
VAULT_SERVER_TLS_CA_CERT_FILE="${VAULT_SERVER_CA_DIR}/ca.cert"
VAULT_SERVER_TLS_CSR_FILE="${VAULT_SERVER_CA_DIR}/vault_server.csr"

# CONSTANTS
VAULT_PROD_WARNING_FILE="/tmp/.olcne-vault-bootstrap-warning"
VAULT_SYSTEMD_FILE="/etc/systemd/system/vault.service"
VAULT_CONFIG_FILE="/etc/vault.d/vault.hcl"
# This file will contain the output of the 'vault operator init' cmd. (root token and unseal keys)
VAULT_INIT_FILE="/etc/vault.d/init.file"

# shellcheck disable=SC2016
SYSTEMD_CONFIG='[Unit]
Description="HashiCorp Vault - A tool for managing secrets"
Documentation=https://www.vaultproject.io/docs/
Requires=network-online.target
After=network-online.target
ConditionFileNotEmpty='${VAULT_CONFIG_FILE}'

[Service]
User=vault
Group=vault
ProtectSystem=full
ProtectHome=read-only
PrivateTmp=yes
PrivateDevices=yes
SecureBits=keep-caps
AmbientCapabilities=CAP_IPC_LOCK
Capabilities=CAP_IPC_LOCK+ep
CapabilityBoundingSet=CAP_SYSLOG CAP_IPC_LOCK
NoNewPrivileges=yes
ExecStart=/usr/bin/vault server -config='${VAULT_CONFIG_FILE}'
ExecReload=/bin/kill --signal HUP $MAINPID
KillMode=process
KillSignal=SIGINT
Restart=on-failure
RestartSec=5
TimeoutStopSec=30
StartLimitBurst=3
LimitNOFILE=65536

[Install]
WantedBy=multi-user.target'


VAULT_OPENSSL_CONFIG='
[req]
default_bits = 2048
encrypt_key  = no # Change to encrypt the private key using des3 or similar
default_md   = sha256
prompt       = no
utf8         = yes
# Speify the DN here so we are not prompted (along with prompt = no above).
distinguished_name = req_distinguished_name
# Extensions for SAN IP and SAN DNS
req_extensions = v3_req
# Be sure to update the subject to match your organization.
[req_distinguished_name]
C=US
ST=California
L="Redwood City"
O="Oracle America"
OU="Private Cloud"
CN = My Certificate
# Allow client and server auth. You may want to only allow server auth.
# Link to SAN names.
[v3_req]
basicConstraints     = CA:FALSE
subjectKeyIdentifier = hash
keyUsage             = digitalSignature, keyEncipherment
extendedKeyUsage     = clientAuth, serverAuth
subjectAltName       = @alt_names
# Alternative names are specified as IP.# and DNS.# for IP addresses and
# DNS accordingly.
[ alt_names ]
'

# Variables defined in 'init'
VAULT_CONFIG_DEV=""
VAULT_CONFIG_SECURE_ADDITONS=""
VAULT_CONFIG_SECURE=""

# Define global variables that rely on user inputs
function init {
    VAULT_CONFIG_DEV='{
    "listener": {
        "tcp": {
            "address": "[::]:8200",
            "cluster_address": "[::]:8201",
            "tls_cipher_suites": "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA,TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA,TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA,TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA,TLS_RSA_WITH_AES_128_GCM_SHA256,TLS_RSA_WITH_AES_256_GCM_SHA384,TLS_RSA_WITH_AES_128_CBC_SHA,TLS_RSA_WITH_AES_256_CBC_SHA",
            "tls_disable": true,
            "tls_prefer_server_cipher_suites": true
        }
    },
    "storage": {
        "file": {
            "path": "'${VAULT_DATA_DIR}'"
        }
    },
    "ui": true
}'

    VAULT_CONFIG_SECURE_ADDITONS='{
    "listener": {
        "tcp": {
            "tls_disable": false,
            "tls_cert_file": "'${VAULT_SERVER_TLS_CERT_FILE}'",
            "tls_key_file": "'${VAULT_SERVER_TLS_KEY_FILE}'"
        }
    }
}'
    VAULT_CONFIG_SECURE=$(jq -s '.[0] * .[1]' <(echo "${VAULT_CONFIG_DEV}") <(echo "${VAULT_CONFIG_SECURE_ADDITONS}"))
}

function usage {
	echo "The Platform Bootstrap tool is used to configure Vault for the Oracle Linux Cloud Native Environment" >&2
    echo ""
    echo "   -a|--vault-addr <VAULT_ADDRESS>"
    echo "      Specify the address for Vault. Mostly used along with '--production' to setup a vault. DEFAULT: ${VAULT_ADDR}"
    echo "   -D|--vault-data-directory <PATH_TO_DIRECTORY>"
    echo "      The directory for the vault data to be stored in. If used with the '--production' flag this value will be ignored. DEFAULT: ${VAULT_DATA_DIR}"
    echo "   -c|--vault-tls-cert-file <PATH_TO_FILE>"
    echo "      Contains the path to the TLS Certificate for the Vault server. If used with the '--production' flag this value will be ignored."
	echo "   -k|--vault-tls-key-file <PATH_TO_FILE>"
    echo "      Contains the path to the TLS Private Key for the Vault server. If used with the '--production' flag this value will be ignored."
	echo "   -e|--extra-vault-tls-cert-sans <EXTRA_SANS>"
    echo "      If the vault TLS cert/key are create by this tool the specified IPs/DNS will be attached to the cert. If used with"
    echo "      the '--production' flag this value will be ignored. DELIMETER: ',' DEFAULT: ${VAULT_SERVER_EXTRA_CERT_SANS}"
	echo "   -s|--vault-tls-skip-verify"
    echo "      Do not verify Vault's presented certificate before communicating with it. Setting this flag is not recommended and voids Vault's security model."
	echo "   -p|--production"
    echo "      This will skip the initialization and configuring of Vault with systemd. We assume you have the pki secrets engine enabled and configued under 'pki/'"
	echo "   -d|--debug"
    echo "      Set up bash debuging"
	echo "   -h|--help"
    echo "      Show the help message"
	echo "usage: " >&2
	echo "   $0 [ --vault-addr <VAULT_ADDRESS> -D|--vault-data-directory <PATH_TO_DIRECTORY> --vault-tls-cert-file <PATH_TO_FILE> --vault-tls-key-file <PATH_TO_FILE> --extra-vault-tls-cert-sans <EXTRA_SANS> --production --debug --help ]" >&2
	exit 0
}

# Update the contents of the file otherwise show warning if we are running in production mode
function update_file {
    file=$1
    expected_file_contents=$2

    diff --ignore-all-space "${file}" <(echo "${expected_file_contents}") > /dev/null 2>&1
    local return_code=$?
    if [[ ${return_code} -ne 0 ]] ; then
        if [ "${PRODUCTION}" == "false" ]; then
            echo "${expected_file_contents}" > "${file}"
        else
            echo -e "[WARNING] Detected ${file} file changed but we are NOT updating. Run without '-p|--production' to overwrite. Expected \n------------\n${expected_file_contents}\n------------"
            return
        fi
    fi
}

function get_port {
    url=$1

    # Get the port from a domain even if it has a path after it. e.g. https://www.oracle.com:8000/test/url
    echo "${url}" | cut -d':' -f3 | grep -o -E '[0-9]+' | head -1
}

function check_exit_code() {
    exit_code=$1
    exit_error=$2

    if [[ ${exit_code} -ne 0 ]]; then
        echo "[ERROR] ${exit_error}"
        exit "${exit_code}"
    fi
}

function wait_for_vault_to_start {
    vault_health_endpoint="${VAULT_ADDR}/sys/health"

    for _ in {1..10}; do
        curl -k "${vault_health_endpoint}" > /dev/null 2>&1
        return_code=$?
        if [ ${return_code} -eq 0 ]; then
            echo "[INFO] Validated Vault is running from: ${vault_health_endpoint}"
            return
        fi

        sleep 5
    done

    echo "[ERROR] Failed to validate vault is running. Response from: ${vault_health_endpoint}"
    curl -k "${vault_health_endpoint}"
    exit 1
}

function check_required_packages {
    # Validate nc is on the system
    which nc 2> /dev/null
    return_code=$?
    if [ ${return_code} -ne 0 ]; then
        echo "[ERROR] The rpm for 'nc' is requred for this tool to run"
        exit 1
    fi

    # Validate jq is on the system
    which jq 2> /dev/null
    return_code=$?
    if [ ${return_code} -ne 0 ]; then
        echo "[ERROR] The rpm for 'jq' is requred for this tool to run"
        exit 1
    fi

    # Validate vault is on the system
    which vault 2> /dev/null
    return_code=$?
    if [ ${return_code} -ne 0 ]; then
        echo "[ERROR] The rpm for 'vault' is requred for this tool to run"
        exit 1
    fi

    # Validate openssl is on the system
    which openssl 2> /dev/null
    return_code=$?
    if [ ${return_code} -ne 0 ]; then
        echo "[ERROR] The rpm for 'openssl' is requred for this tool to run"
        exit 1
    fi
}

function validate_ip {
    local  ip=$1
    local  ret=1

    # check the given ip contains exactly 4 octets
    if [[ $ip =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$ ]]; then
        #lets break the ip into octets array
        prev_ifs=$IFS
        IFS='.'
        # shellcheck disable=SC2206
        ip=(${ip})
        #lets restore IFS once we are done with it
        IFS=$prev_ifs
        #check each octet is <=255
        [[ ${ip[0]} -le 255 && ${ip[1]} -le 255 \
            && ${ip[2]} -le 255 && ${ip[3]} -le 255 ]]
        ret=$?
    fi

    return $ret
}

function create_certs {
    cert_sans_values=$1

    if [[ -f ${VAULT_SERVER_TLS_CERT_FILE} && -f ${VAULT_SERVER_TLS_KEY_FILE} ]]; then
        echo "[INFO] Skipping generating certs because the Vault Server key and cert exist"
        return
    fi

    # Validate we have both cert and key gone before continuing
    if [[ -f ${VAULT_SERVER_TLS_CERT_FILE} || -f ${VAULT_SERVER_TLS_KEY_FILE} ]]; then
        echo "[ERROR] Can't generate certs when only one CERT/KEY exists. Please validate both"
        echo "        CERT and KEY exist or are cleaned up before continuing."
        echo "        CERT: ${VAULT_SERVER_TLS_CERT_FILE}"
        echo "        KEY:  ${VAULT_SERVER_TLS_KEY_FILE}"
        exit 1
    fi

    VAULT_SERVER_TLS_CSR_DETAILS=$(mktemp)

    # Assemble cert sans
    num_dns_alt_names=1
    num_ip_alt_names=1
    alt_names_str=""

    prev_ifs=$IFS
    IFS=","
    for master_ip in ${cert_sans_values}; do
        if validate_ip "$master_ip"; then
            alt_names_str="$alt_names_str\nIP.${num_ip_alt_names} = ${master_ip}"
            ((num_ip_alt_names++))
        else
            alt_names_str="$alt_names_str\nDNS.${num_dns_alt_names} = ${master_ip}"
            ((num_dns_alt_names++))
        fi
    done
    IFS=$prev_ifs

    echo -e "${VAULT_OPENSSL_CONFIG}${alt_names_str}" > "${VAULT_SERVER_TLS_CSR_DETAILS}"

    openssl req \
        -new \
        -newkey rsa:2048 \
        -days 3650 \
        -nodes \
        -x509 \
        -subj "/C=US/ST=California/L=Redwood City/O=Oracle America CA" \
        -keyout "${VAULT_SERVER_TLS_CA_KEY_FILE}" \
        -out "${VAULT_SERVER_TLS_CA_CERT_FILE}"

    openssl genrsa -out "${VAULT_SERVER_TLS_KEY_FILE}" 2048

    openssl req \
        -new -key "${VAULT_SERVER_TLS_KEY_FILE}" \
        -out "${VAULT_SERVER_TLS_CSR_FILE}" \
        -config "${VAULT_SERVER_TLS_CSR_DETAILS}"

    openssl x509 \
        -req \
        -days 3650 \
        -in "${VAULT_SERVER_TLS_CSR_FILE}" \
        -CA "${VAULT_SERVER_TLS_CA_CERT_FILE}" \
        -CAkey "${VAULT_SERVER_TLS_CA_KEY_FILE}" \
        -CAcreateserial \
        -extensions v3_req \
        -extfile "${VAULT_SERVER_TLS_CSR_DETAILS}" \
        -out "${VAULT_SERVER_TLS_CERT_FILE}" > /dev/null

    rm "${VAULT_SERVER_TLS_CSR_DETAILS}"
}

function vault_configure_system {
    echo "[INFO] Configuring systemd"

    # Check if the file changed
    function file_change_check {
    update_file "${VAULT_SYSTEMD_FILE}" "${SYSTEMD_CONFIG}"
    update_file "${VAULT_CONFIG_FILE}" "${VAULT_CONFIG_SECURE}"
    }


    # Nothing past this point is required for a production env
    if [ "${PRODUCTION}" == "true" ]; then
	file_change_check
        return
    fi

    # Check if vault user exists
    id -u vault > /dev/null
    return_code=$?
    if [[ ${return_code} -ne 0 ]]; then
        # Create a unique, non-privileged system user to run Vault.
        sudo useradd --system --home /etc/vault.d --shell /bin/false vault
        check_exit_code "$?" "creating the vault user"
    fi

    # Prevents memory from being swapped to disk via mlock
    sudo setcap cap_ipc_lock=+ep /usr/bin/vault
    check_exit_code "$?" "disabling memmory from being swapped to disk for vault"

    mkdir -p "${VAULT_SERVER_CA_DIR}"

    create_certs ${VAULT_SERVER_EXTRA_CERT_SANS}

    sudo mkdir -p ${VAULT_DATA_DIR}
    sudo chown --recursive vault:vault ${VAULT_DATA_DIR}

    # Create Vault config dir
    sudo mkdir --parents /etc/vault.d
    sudo touch ${VAULT_CONFIG_FILE}
    sudo chown --recursive vault:vault /etc/vault.d
    sudo chown --recursive vault:vault /var/lib/vault.d

    # Reset and Restart Vault
    sudo rm -rf  "${VAULT_DATA_DIR}"
    sudo -u vault mkdir -p "${VAULT_DATA_DIR}"

    sudo chmod 640 ${VAULT_CONFIG_FILE}

    file_change_check

    systemctl enable vault
    systemctl restart vault

    wait_for_vault_to_start
}

function vault_init {
    echo "[INFO] Initilizing Vault"
    # Don't setup vault if its a production environement
    if [ "${PRODUCTION}" == "true" ]; then
        echo "[INFO] Skipping the initilizion of Vault for the production environement"
        return
    fi

    vault operator init > ${VAULT_INIT_FILE}
    check_exit_code "$?" "running 'vault operator init'. Look into running with '--vault-tls-skip-verify' if your seeing an 'unknown authority' error"

    # Only add ports if the firewalld is running
    firewall-cmd --state > /dev/null 2>&1
    return_code=$?
    if [ ${return_code} -eq 0 ]; then
        vault_port=$(get_port "${VAULT_ADDR}")
        if [ "${vault_port}" != "" ]; then
            firewall-cmd --add-port "${vault_port}/tcp" --permanent > /dev/null 2>&1
            firewall-cmd --add-port "${vault_port}/tcp" > /dev/null 2>&1
        fi
    fi

    ROOT_TOKEN=$(grep "Initial Root Token" ${VAULT_INIT_FILE} | cut -d':' -f2 | awk '{ print $1 }')

    # Unseal vault on the startup
    grep "Unseal Key" ${VAULT_INIT_FILE} | cut -d':' -f2 | awk '{ print $1 }' | xargs -n1 vault operator unseal

    vault login "${ROOT_TOKEN}"

    vault secrets enable pki
    check_exit_code "$?" "enabling the pki secrets engine at the pki path"

    vault secrets tune -max-lease-ttl=87600h pki
    check_exit_code "$?" "tuning the pki secrets engine to issue certificates with a maximum time-to-live (TTL) of 87600 hours."

    vault write -field=certificate pki/root/generate/internal \
        common_name="olcne" \
        ttl=87600h > CA_cert.crt
    check_exit_code "$?" "creating a root cert"

    vault write pki/config/urls \
        issuing_certificates="${VAULT_ADDR}/v1/pki/ca" \
        crl_distribution_points="${VAULT_ADDR}/v1/pki/crl"
    check_exit_code "$?" "configuring the CA and CRL URLs with ${VAULT_ADDR}"
}

# Generate Intermediate CA for OLCNE
function vault_olcne_setup {
    output=$(vault status 2>&1)
    return_code=$?
    if [[ ${return_code} -ne 0 ]]; then
        echo "[ERROR] Failed to validate Vault cli is configured with 'vault status'. Common issues are:"
        echo "    - Verify Vault is unsealed"
        echo "    - Verify the firewall port is open to talk to the Platform API Server"
        echo "${output}"
        exit 1
    fi

    vault secrets enable -path=olcne_pki_intermediary pki
    check_exit_code "$?" "enabling olcne_pki_intermediary vault pki secret"

    vault secrets tune -max-lease-ttl=87600h olcne_pki_intermediary
    check_exit_code "$?" "tuning the olcne_pki_intermediary secrets engine to issue certificates with a maximum time-to-live (TTL) of 10 years."

    # Generate an intermediate and save the CSR as pki_intermediate.csr:
    vault write -format=json olcne_pki_intermediary/intermediate/generate/internal \
        common_name="olcne Intermediate Authority" ttl="87600h" \
        | jq -r '.data.csr' > pki_intermediate.csr
    check_exit_code "$?" "generating the pki csr"

    vault write -format=json pki/root/sign-intermediate csr=@pki_intermediate.csr \
        format=pem_bundle ttl="87600h" \
        | jq -r '.data.certificate' > intermediate.cert.pem
    check_exit_code "$?" "signing the intermediate certificate with the root certificate and save the generated certificate as intermediate.cert.pem"

    # Once the CSR is signed and the root CA returns a certificate, it can be imported back into Vault:
    vault write olcne_pki_intermediary/intermediate/set-signed certificate=@intermediate.cert.pem
    check_exit_code "$?" "importing the intermidiate cert back into vault"

    # Remove the intermediate files
    rm -f pki_intermediate.csr
    rm -f intermediate.cert.pem

    # Create a role named olcne which allows any certificate. Set 'use_csr_common_name' 'require_cn' and 'require_cn' so
    # we can specify alt_names and ip_sans when creating x509 from the agent
    vault write olcne_pki_intermediary/roles/olcne \
        allow_any_name=true \
        use_csr_common_name=false \
        require_cn=false \
        use_csr_san=false \
        max_ttl="87600h"
    check_exit_code "$?" "creating a role for the Platform PKI intermediary"
}

function main {
    return_code=0

    init

    # Export the Vault address for the cli
    export VAULT_ADDR

    # Show warning to users running against a production Vault
    if [ ! -f "${VAULT_PROD_WARNING_FILE}" ] && [ "${PRODUCTION}" == "true" ] ; then
        echo "[WARNING] If you have a production Vault use this script with causion. We are generating"
        echo "  root certificates, configuring Vault under systemd and creating policies. This message will"
        echo "  only show up ONCE."
        touch ${VAULT_PROD_WARNING_FILE}
        exit
    fi

    echo "[INFO] Configuring Vault"

    # Validate required packages are on the system
    check_required_packages

    # Configure the system for vault, create vault user and setup systemd - no writes are done for production
    vault_configure_system

    # Initialize Vault, configure the root CA and setup the firewall - skipped for production
    vault_init

    # Setup the olcne part of vault
    vault_olcne_setup

        echo '
    path "olcne_pki_intermediary/*" {
        capabilities = [ "update" ]
    }
    ' > olcne-pki-policy.hcl

    vault policy write olcne-agent olcne-pki-policy.hcl
    olcne_access_token=$(vault token create -policy=olcne-agent --format=json | jq '.auth.client_token' | sed 's/"//g')
    echo -e "\n\nAccess Token: ${olcne_access_token}" >> ${VAULT_INIT_FILE}

    echo "[SUCCESS] Vault is ready"
    echo ""
    echo "[DATA] Vault Root Token and Unseal Keys:"
    echo "    ${VAULT_INIT_FILE}"
    echo "[DATA] Vault Platform Access Token:"
    echo "    ${olcne_access_token}"
}

# MAIN
while [[ $# -gt 0 ]]; do
key="$1"

case $key in
    -a | --vault-addr )
        VAULT_ADDR="$2"
        shift # shift past argument
        shift # shift past value
    ;;
    -D | --vault-data-directory )
        VAULT_DATA_DIR="$2"
        shift # shift past argument
        shift # shift past value
    ;;
    -p | --production )
        PRODUCTION="true"
        shift # shift past argument
    ;;
   -c | --vault-tls-cert-file )
        VAULT_SERVER_TLS_CERT_FILE="$2"
        shift # shift past argument
        shift # shift past value
    ;;
   -k | --vault-tls-key-file )
        VAULT_SERVER_TLS_KEY_FILE="$2"
        shift # shift past argument
        shift # shift past value
    ;;
   -e | --extra-vault-tls-cert-sans )
        VAULT_SERVER_EXTRA_CERT_SANS="$2"
        shift # shift past argument
        shift # shift past value
    ;;
   -s | --vault-tls-skip-verify )
        export VAULT_SKIP_VERIFY="true"
        shift # shift past argument
    ;;
   -d | --debug )
        set -x
        shift # shift past argument
    ;;
    -h | --help )
        usage
        exit
    ;;
    *) # unknown option
        echo "[ERROR] unknown argument: ${key}"
        usage
        exit 1
    ;;
esac
done

echo "VAULT_ADDR: $VAULT_ADDR PRODUCTION: $PRODUCTION VAULT_SERVER_TLS_CERT_FILE: $VAULT_SERVER_TLS_CERT_FILE VAULT_SERVER_TLS_KEY_FILE: $VAULT_SERVER_TLS_KEY_FILE"
main
